# -*- coding: utf-8 -*-

"""

  setup tests
  ~~~~~~~~~~~

  makes sure canteen's setup routine works properly.

  :author: Sam Gammon <sg@samgammon.com>
  :copyright: (c) Sam Gammon, 2014
  :license: This software makes use of the MIT Open Source License.
            A copy of this license is included as ``LICENSE.md`` in
            the root of the project.

"""

# setup
from canteen import setup
