# -*- coding: utf-8 -*-

"""

  HTTP logic tests
  ~~~~~~~~~~~~~~~~

  :author: Sam Gammon <sg@samgammon.com>
  :copyright: (c) Sam Gammon, 2014
  :license: This software makes use of the MIT Open Source License.
            A copy of this license is included as ``LICENSE.md`` in
            the root of the project.

"""

if __debug__:

  __all__ = ('test_agent',
             'test_caching',
             'test_cookies',
             'test_etags',
             'test_redirects',
             'test_semantics')
