# -*- coding: utf-8 -*-

"""

  core injection tests
  ~~~~~~~~~~~~~~~~~~~~

  tests the injection core, which is the "frontend" to the DI system.
  "compound" meta-objects, defined by the module tested by this code,
  can make use of the attribute injection pool locally via MRO.

  :author: Sam Gammon <sg@samgammon.com>
  :copyright: (c) Sam Gammon, 2014
  :license: This software makes use of the MIT Open Source License.
            A copy of this license is included as ``LICENSE.md`` in
            the root of the project.

"""
